/* David Leonard, 2002. Public domain. */

/*
 * ActiveSync(tm) keep-alive daemon
 */

#include <netdb.h>
#include <stdio.h>
#include <unistd.h>
#include <syslog.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <netinet/in.h>
#include <machine/endian.h>
#include <arpa/inet.h>

#include "ustring.h"
#include "syslogerr.h"

#define ACTIVESYNC_PORT	5679
#define PING		0x12345678
#define PING_INTERVAL	5000		/* in milliseconds */

/* Dump a block of data for debugging purposes */
void
dump(desc, data, len)
	const char *desc;
	void *data;
	size_t len;
{
	u_int8_t *buf = data;
	int i, j;
	char hex[8 * 3 + 1];
	char chr[8 + 1];

	syslog(LOG_DEBUG, "%s (%d bytes):", desc, len);
	for (i = 0; i < len + 7; i += 8) {
	    for (j = 0; j < 8; j++) 
		if (j + i >= len) {
			hex[3*j+0] = ' ';
			hex[3*j+1] = ' ';
			hex[3*j+2] = ' ';
			chr[j] = ' ';
		} else {
			u_int8_t c = buf[j + i];
			const char *hexchr = "0123456789abcdef";
			hex[3*j+0] = hexchr[(c >> 4) & 0xf];
			hex[3*j+1] = hexchr[c & 0xf];
			hex[3*j+2] = ' ';
			if (c > ' ' && c <= '~')
				chr[j] = c;
			else
				chr[j] = '.';
		}
	    hex[8*3] = '\0';
	    chr[8] = '\0';
	    syslog(LOG_DEBUG, "  %04x: %s %s", i, hex, chr);
	}
}

/* 'Encrypt' a passphrase into the given buffer. Returns size of buffer used */
size_t
xorcrypt(pass, buf, bufsiz, key)
	const char *pass;
	void *buf;
	size_t bufsiz;
	u_int8_t key;
{
	int i;
	size_t size;

	size = ctou((uchar *)buf, bufsiz, pass);
	for (i = 0; i < size; i++)
		((unsigned char *)buf)[i] ^= key;
	return size;
}


int
main(argc, argv)
	int argc;
	char *argv[];
{
	extern int optind;
	extern char *optarg;
	int ch;
	int error = 0;
	int lflag = 0;
	int len;
	socklen_t socklen;
	int pingsmissed = 0;
	char *passphrase = NULL;
	char devname[128], devclass[128], devhw[128];
	struct sockaddr_in from;

	/* Process command line arguments */
	while ((ch = getopt(argc, argv, "lp:")) != -1)
		switch (ch) {
		case 'l':
			lflag = 1;
			break;
		case 'p':
			passphrase = optarg;
			break;
		default:
			error = 1;
		}
	if (argc != optind)
		error = 1;

	/* -l option also logs to stderr */
	openlog("asyncd", LOG_PID | (lflag ? LOG_PERROR : 0), LOG_DAEMON);

	if (error)
		errx(1, "usage: %s [-l] [-p passphrase]", argv[0]);

	/* -l flag listens for and accepts one connection */
	if (lflag) {
		int s, t;
		struct sockaddr_in saddr;
		struct linger l;

		saddr.sin_family = AF_INET;
		saddr.sin_port = htons(ACTIVESYNC_PORT);
		saddr.sin_addr.s_addr = INADDR_ANY;
		socklen = sizeof saddr;
		if ((s = socket(PF_INET, SOCK_STREAM, 0)) == -1)
			err(1, "socket");
		if (bind(s, (struct sockaddr *)&saddr, socklen) == -1)
			err(1, "bind");
		l.l_onoff = 1;
		l.l_linger = 1;
		if (setsockopt(s, SOL_SOCKET, SO_LINGER, &l, sizeof l) == -1)
			warn("SO_LINGER");
		if (listen(s, 1) == -1)
			err(1, "listen");
		syslog(LOG_DEBUG, "waiting for connection");
		if ((t = accept(s, NULL, NULL)) == -1)
			err(1, "accept");
		close(s);

		/* Simulate being run from inetd */
		dup2(t, STDIN_FILENO);
		dup2(t, STDOUT_FILENO);
		close(t);
	}

	/* Determine where the connection is from */
	socklen = sizeof from;
	if (getpeername(STDIN_FILENO, (struct sockaddr *)&from, &socklen) == -1)
		err(1, "getpeername");
	syslog(LOG_INFO, "connection from %s:%u",
		inet_ntoa(from.sin_addr),
		ntohs(from.sin_port));

	/* Process messages on the connection */
	for (;;) {
		u_int32_t u32, header;
		struct pollfd pfd;
		u_int8_t buf[512];
		int off;

		pfd.fd = STDIN_FILENO;
		pfd.events = POLLIN;
		pfd.revents = 0;

		/* Wait for a message */
		if (poll(&pfd, 1, PING_INTERVAL) == -1)
			err(1, "poll");

		/* Handle timeout by sending a ping */
		if ((pfd.revents & POLLIN) == 0) {
			pingsmissed++;
			if (pingsmissed > 4) {
				syslog(LOG_DEBUG, "too many pings lost");
				break;
			}
			syslog(LOG_DEBUG, "sending ping");
			u32 = htole32(PING);
			if (write(STDOUT_FILENO, &u32, sizeof u32) == -1)
				err(1, "write");
			continue;
		}

		/* Read message header */
		if ((len = read(STDIN_FILENO, &u32, sizeof u32)) == -1) {
			if (errno == ECONNRESET) {
				syslog(LOG_INFO, "%m");
				break;
			}
			err(1, "read");
		}
		if (len == 0)
			break;
		header = letoh32(u32);
		syslog(LOG_DEBUG, "receive 0x%08x", header);

		/* Ignore empty messages */
		if (header == 0)
			continue;

		/* Receive ping reply */
		if (header == PING) {
			syslog(LOG_DEBUG, "heard ping reply");
			pingsmissed = 0;
			continue;
		}

		/* Handle challenge request */
		if (header > 0x10000) {
			struct {
				u_int16_t len;
				char crypt[512];
			} response;
			u_int8_t challengekey;
			size_t len;

			if (passphrase == NULL) {
				/* XXX use getpass(3)? */
				syslog(LOG_DEBUG, 
				    "challenge request but no -p option");
				break;
			}
			challengekey = header & 0xff;
			syslog(LOG_DEBUG, "challenge 0x%02x", challengekey);

			len = xorcrypt(passphrase, response.crypt,
			    sizeof response.crypt, challengekey);
			response.len = htole16(len);
			len += sizeof (u_int16_t);

			dump("sending response", &response, len);

			if (write(STDOUT_FILENO, &response, len) == -1)
				err(1, "write");
			continue;
		}

		/* Handle information message */
		if (header > sizeof buf)
			err(1, "packet too big");
		if ((len = read(STDIN_FILENO, buf, header)) == -1)
			err(1, "read");

		if (len == 0)
			errx(1, "connection closed");

		dump("received info", buf, header);

		/*
		 *	The info packet dump I get from my iPAQ is:
		 *	0000:	0x00000024
		 *	0004:	03 00 84 24
		 *	0008:	11 0a 00 00
		 *	000c:	00 00 00 00
		 *	0010:	ba 03 62 47
		 *	0014:	0e 33 2d 68
		 *	0018:	0x00000024		- (offset to name)
		 *	001c:	0x00000032		- (offset to class)
		 *	0020:	0x00000044		- (offset to hw)
		 *	0024:	L"thingy\0"		- name
		 *	0032:	L"Palm PC2\0"		- class
		 *	0044:	L"Compaq iPAQ H3600\0"	- hw
		 */
		
		/* Extract device information */
		off = letoh32(*(u_int32_t *)(buf + 0x18));
		utoc(devname, sizeof devname, (uchar *)(buf + off));
		off = letoh32(*(u_int32_t *)(buf + 0x1c));
		utoc(devclass, sizeof devclass, (uchar *)(buf + off));
		off = letoh32(*(u_int32_t *)(buf + 0x20));
		utoc(devhw, sizeof devhw, (uchar *)(buf + off));

		syslog(LOG_INFO, "name \"%s\" class \"%s\" hw \"%s\"",
			devname, devclass, devhw);

		/* XXX could run a shell script here to start syncing */

	}
	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	syslog(LOG_INFO, "connection closed");
	exit(0);
}
